/*
 *  Print Pango Font Families. (family name, family's face)
 *
 *      $ gcc `pkg-config gtk+-2.0 --cflags --libs` print_pango_font.c -o pango_font
 *      $ ./pango_font
 */
#include <gtk/gtk.h>

void PrintPangoFontFamilies(GtkWidget *widget)
{
    PangoContext *pango_context;
    PangoFontFamily **families;
    int n_families;
    int i, j;

    pango_context = gtk_widget_get_pango_context(widget);
    pango_context_list_families (pango_context, &families, &n_families);

    printf("\n");

    for (i = 0; i < n_families; i++)
    {
        gchar *family = (gchar *)pango_font_family_get_name(families[i]);
        printf("Font family: %s\n", family);

        PangoFontFace **faces;
        int n_faces;

        pango_font_family_list_faces(families[i], &faces, &n_faces);
        printf("       face:");

        for (j = 0; j < n_faces; j++)
        {
            gchar *face = (gchar *)pango_font_face_get_face_name (faces[j]);
            if (j < n_faces -1)
            {
                printf(" %s,", face);
            }
            else
            {
                printf(" %s", face);
            }
        }
        printf("\n\n");
        g_free(faces);
   }

    g_free (families);
}

int main (int argc, char *argv[])
{
    GtkWidget *window;
    gtk_init (&argc, &argv);
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

    // Print Pango Font Families. (family name, family's face)
    PrintPangoFontFamilies(window);

    return 0;
}